library mt4R;

{$mode objfpc}{$H+}

uses
  cmem, Classes, Windows, sysutils, profs_TRConsole, profs_win32exceptiontrap;


{internal helper functions}

// log a message to the debug monitor
procedure Log(AMessage: String);
begin
  OutputDebugString(PChar(AMessage));
end;

// log a formatted message to the debug monitor
procedure Log(AMessage: String; AArgs: array of const);
begin
  Log(Format(AMessage, AArgs));
end;

// return true if the R session object belonging to the handle is valid
function isValid(AHandle: LongInt): Boolean;
var
  Dummy : Boolean;
begin
  // the ea might be so messed up that it called RDeinit once too often,
  // so we will always wrap this into AddHandler/RemoveHandler
  AddHandler;
  try
    // cast the "handle" to the pointer that it is in reality
    // and then just try to access some field
    Dummy := TRConsole(AHandle).Running;
    Result := True;
  except
    // if invalid this should have raised an access violation
    Log('Invalid (not existing) handle for TRConsole object: %d', [AHandle]);
    Dummy := Dummy; // suppress the compiler hint about unused variable
    Result := False;
  end;
  RemoveHandler;
end;


{ exported functions}

// start a new R session and return the handle. The "handle" is actually
// a pointer cast to a LongInt (32 bit) and in mql4 it can be treated like
// a handle. The other functions will simply cast it back into TRConsole.
function RInit(ACommandLine: PChar; ADebugLevel: LongInt): LongInt; stdcall;
var
  R: TRConsole;
begin
  AddHandler;
  R := TRConsole.Create(ACommandLine, ADebugLevel);
  Result := LongInt(R);
  Log('RInit: RHandle = %x (%d)', [Result, Result]);
end;

// terminate the R session
procedure RDeinit(AHandle: LongInt); stdcall;
begin
  Log('RDeinit: RHandle = %x (%d)', [AHandle, AHandle]);
  if isValid(AHandle) then begin
    TRConsole(AHandle).Free;
  end;
  RemoveHandler;
end;

// execute code and wait
procedure RExecute(AHandle: LongInt; ACode: PChar); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).ExecuteCode(ACode);
  end;
end;

// execute code and wait, same as above but write to temp file and source()
procedure RExecuteIndirect(AHandle: LongInt; ACode: PChar); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).ExecuteCodeIndirect(ACode);
  end;
end;

// assign integer to variable given by name
procedure RAssignInteger(AHandle: LongInt; AVariable: PChar; AValue: LongInt); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignInteger(AVariable, AValue);
  end;
end;

// assign double to variable given by name
procedure RAssignDouble(AHandle: LongInt; AVariable: PChar; AValue: Double); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignDouble(AVariable, AValue);
  end;
end;

// assign string to variable given by name
procedure RAssignString(AHandle: LongInt; AVariable: PChar; AValue: PChar); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignString(AVariable, AValue);
  end;
end;

// assign vector to variable given by name
procedure RAssignVector(AHandle: LongInt; AVariable: PChar; AVector: PVector; ASize: LongInt); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignVector(AVariable, AVector, ASize);
  end;
end;

// assign factor to variable given by name
procedure RAssignFactor(AHandle: LongInt; AVariable: PChar; AFactor: PStrVector; ASize: LongInt); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignFactor(AVariable, AFactor, ASize);
  end;
end;

// assign a matrix to the variable give by name
procedure RAssignMatrix(AHandle: LongInt; AVariable: PChar; AMatrix: PVector; ARows: LongInt; ACols: LongInt); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).AssignMatrix(AVariable, AMatrix, ARows, ACols);
  end;
end;

// variable <- rbind(variable, vector)
procedure RRowBindVector(AHandle: LongInt; AVariable: PChar; AVector: PVector; ASize: LongINt); stdcall;
begin
  if isValid(AHandle) then begin
    TRConsole(AHandle).RowBindVector(AVariable, AVector, ASize);
  end;
end;

// evaluate expression and return integer
function RGetInteger(AHandle: LongInt; AExpression: PChar): LongInt; stdcall;
begin
  if isValid(AHandle) then
    Result := TRConsole(AHandle).GetInteger(AExpression)
  else
    Result := 0;
end;

// evaluate expression and return double
function RGetDouble(AHandle: LongInt; AExpression: PChar): Double; stdcall;
begin
  if isValid(AHandle) then
    Result := TRConsole(AHandle).GetDouble(AExpression)
  else
    Result := 0;
end;

// evaluate expression and return vector into supplied array.
// Return value is the actual number of elements that have been copied.
// if the size of the array does not match the size of the actuall vector
// then a warning will be emitted on debuglevel 1. Return value will
// always be equal or smaller than the supplied ASize.
function RGetVector(AHandle: LongInt; AExpression: PChar; AVector: PVector; ASize: LongInt): LongInt; stdcall;
begin
  if isValid(AHandle) then
    Result := TRConsole(AHandle).GetVector(AExpression, AVector, ASize)
  else
    Result := 0;
end;

procedure  RDebug(AHandle: LongInt; AVariable: PChar); stdcall;
begin
  if isValid(AHandle) then
    TRConsole(AHandle).Debug(AVariable);
end;

exports
  RInit,
  RDeinit,
  RExecute,
  RExecuteIndirect,
  RAssignInteger,
  RAssignDouble,
  RAssignString,
  RAssignVector,
  RAssignFactor,
  RAssignMatrix,
  RRowBindVector,
  RGetInteger,
  RGetDouble,
  RGetVector,
  RDebug;

{$R *.res}

begin
end.

