unit profs_TRConsole;

{$mode objfpc}{$H+}

interface
uses
  process;

type
  // this is double foo[] in mql4
  TVector = array [0..maxLongint div SizeOf(Double) - 1] of Double;
  PVector = ^TVector;

  TMqlStr = packed record
    Size: LongInt;
    Ptr: PChar;
  end;

  // this is string foo[] in mql4.
  TStrVector = array [0..maxLongint div SizeOf(TMqlStr) - 1] of TMqlStr;
  PStrVector = ^TStrVector;

  { TRconsole }

  TRConsole = class(TProcess)
  protected
    FDebugLevel : Integer;
    procedure Msg(ALevel: Integer; AMethod: AnsiString; AMessage: AnsiString);
    procedure Msg(ALevel: Integer; AMethod: AnsiString; AMessage: AnsiString; AArgs: array of const);
    function GetOutput: AnsiString;
    function ConvertDouble(AStr: AnsiString): Double;
  public
    // start an instance of R (c:\full\path\to\Rterm.exe --no-save)
    constructor Create(ACommandLine: AnsiString; ADebugLevel: LongInt); reintroduce;

    // terminate the R session (don't call this, call Free as usual)
    destructor Destroy; override;

    // DebugLevel: 0=errors, 1=warnings, 2=all
    property DebugLevel: Integer read FDebugLevel write FDebugLevel;

    // format a vector
    function FormatVector(AVector: PVector; ASize: LongInt): AnsiString;

    // output info about the variable on debuglevel 1
    procedure Debug(AVariable: AnsiString);

    // evaluate expression and return the raw output
    function ExecuteCode(ACode: AnsiString): AnsiString;

    // evaluate expression and return the raw output
    // the difference to ExecuteCode is that it will first write
    // the code to a temporary file and then call source()
    function ExecuteCodeIndirect(ACode: AnsiString): AnsiString;

    // assign the integer value to the variable
    procedure AssignInteger(AVariable: AnsiString; AValue: LongInt);

    // assign the double value to the variable
    procedure AssignDouble(AVariable: AnsiString; AValue: Double);

    // assign the string to the variable
    procedure AssignString(AVariable: AnsiString; AValue: AnsiString);

    // assign the vector to the variable
    procedure AssignVector(AVariable: AnsiString; AVector: PVector; ASize: LongInt);

    // assign the factor to the variable
    procedure AssignFactor(AVariable: AnsiString; AFactor: PStrVector; ASize: LongInt);

    // assign the matrix to the variable.
    // AMatrix is a pointer to the raw 2d array of doubles, first dimenstion is row.
    procedure AssignMatrix(AVariable: AnsiString; AMatrix: PVector; ARows: LongInt; ACols: LongInt);

    // append the vector as a new row to the variable
    procedure RowBindVector(AVariable: AnsiString; AVector: PVector; ASize: LongInt);

    // evaluate expression and return an integer
    function GetInteger(ACode: AnsiString): LongInt;

    // evaluate expression and retun a double
    function GetDouble(ACode: AnsiString): Double;

    // evaluate expression and return an array of double
    // and the number of elements that were actually copied.
    // the supplied array must be big enough or it will not copy
    // all elements. The ACode must evaluate to a vector.
    function GetVector(ACode: AnsiString; AVector: PVector; ASize: LongInt): LongInt;

  end;


implementation

uses
  {$ifdef win32}windows,{$endif}
  Classes, SysUtils, strutils, math;

resourcestring
  rsDebugStoppingR = 'Stopping R';
  rsDebugSuccessfulStart = 'R successfully started';
  rsDebugTryingToStartR = 'trying to start R: %s';
  rsErrArraySize = 'Error: array size mismatch';
  rsErrCouldNotConvertDouble = 'Error: Could not convert to double: %s';
  rsErrCouldNotConvertToInt = 'Error: Could not convert to int: %s';
  rsErrDidNotStart = 'Error: R did not start: %d';
  rsErrRNotRunning = 'Error: R is not running (anymore): %s';
  rsErrWriteTemp = 'Error writing to temp file';
  rsNoteArrayBiggerThanNeeded = 'Note: array [%d] was bigger than really needed [%d]';
  rsWarnArrayTooSmall = 'Warning: array [%d] not big enough [%d], vector truncated';
  rsWarnNull = 'Warning: NULL';

constructor TRConsole.Create(ACommandLine : AnsiString; ADebugLevel : LongInt);
begin
  inherited Create(nil);
  FDebugLevel := ADebugLevel;
  CommandLine := ACommandLine;
  Msg(2, 'Create', Format(rsDebugTryingToStartR, [ACommandLine]));
  Options := [poUsePipes, poStderrToOutPut, poNoConsole];

  Execute;
  if Running then begin
    GetOutput;
    ExecuteCode('options(digits=15)');
    Msg(2, 'Create', rsDebugSuccessfulStart);
  end else begin
    Msg(0, 'Create', Format(rsErrDidNotStart, [GetLastOSError]));
  end;
end;

destructor TRconsole.Destroy;
begin
  Msg(2, 'Destroy', rsDebugStoppingR);
  Terminate(0);
  inherited Destroy;
end;

procedure TRconsole.Msg(ALevel: Integer; AMethod: AnsiString; AMessage: AnsiString);
begin
  if ALevel <= FDebugLevel then begin
    AMessage := '<' + IntToStr(ALevel) + '> TRConsole.' + AMethod + ': ' + AMessage;
    {$ifdef win32}
    OutputDebugString(PChar(AMessage));
    {$else}
    writeln(AMessage);
    {$endif}
  end;
end;

procedure TRConsole.Msg(ALevel: Integer; AMethod: AnsiString; AMessage: AnsiString; AArgs: array of const);
begin
  try
    Msg(ALevel, AMethod, Format(AMessage, AArgs));
  except
    Msg(ALevel, AMethod, 'could not format debug message');
  end;
end;

function TRconsole.GetOutput: AnsiString;
const
  RPrompt = LineEnding + '> ';
var
  BytesRead: DWord;
  Position: DWord;
  NumAvail: DWord;
begin
  Result := '';
  Position := 0;
  while True do begin
    NumAvail := Output.NumBytesAvailable;
    if NumAvail = 0 then begin
      Sleep(1);
      if not Running then begin
        Msg(0, 'GetOutput', rsErrRNotRunning, [Result]);
        exit;
      end;
    end else begin
      SetLength(Result, Position + NumAvail);
      BytesRead := Output.Read(Result[Position + 1], NumAvail);
      Position := Position + BytesRead;
      if RightStr(Result, Length(RPrompt)) = RPrompt then begin
        Result := LeftStr(Result, Length(Result) - Length(RPrompt));
        break;
      end;
    end;
  end;
end;

function TRconsole.ConvertDouble(AStr: AnsiString): Double;
begin
  if AStr = 'Inf' then exit(Infinity);
  if AStr = '-Inf' then exit(NegInfinity);
  if AStr = 'NaN' then exit(NaN);
  try
    Result := StrToFloat(AStr);
  except
    Msg(0, 'ConvertDouble', rsErrCouldNotConvertDouble, [AStr]);
    Result := 0;
  end;
end;

function TRConsole.FormatVector(AVector: PVector; ASize: LongInt): AnsiString;
var
  i: LongInt;
begin
  DecimalSeparator := '.';
  Result := 'c(';
  ASize -= 1;
  for i := 0 to ASize do begin
    Result := Result + FloatToStr(AVector^[i]);
    if i < ASize then Result := Result + ',';
  end;
  Result := Result + ')';
end;

procedure TRConsole.Debug(AVariable: AnsiString);
begin
  if DebugLevel > 0 then begin
    Msg(1, 'Debug', 'str(' + AVariable + '): ' + ExecuteCode('str(' + AVariable + ')'));
    Msg(1, 'Debug', 'length(' + AVariable + '): ' + ExecuteCode('length(' + AVariable + ')'));
  end;
end;

function TRconsole.ExecuteCode(ACode: AnsiString): AnsiString;
begin
  Result := '';
  if Running then begin
    Msg(2, 'ExecuteCode', 'in  >>>  ' + ACode);
    ACode := ACode + LineEnding;
    Input.Write(ACode[1], Length(ACode));
    Result := GetOutput;
    Result := RightStr(Result, Length(Result) - Length(ACode)); // remove echo
    Msg(2, 'ExecuteCode', 'out <<<  ' + Result);
  end else begin
    Msg(0, 'ExecuteCode', rsErrRNotRunning, [ACode]);
  end;
end;

function TRConsole.ExecuteCodeIndirect(ACode: AnsiString): AnsiString;
var
  FN : AnsiString;
  F: TFileStream;
begin
  Msg(2, 'ExecuteCodeIndirect', LeftStr(ACode, 40) + '...');
  FN := GetTempFileName(GetTempDir, 'TRConsole-source-');
  FN := StringsReplace(FN, ['\'], ['/'], [rfReplaceAll]);
  try
    F := TFileStream.Create(FN, fmCreate);
    F.WriteBuffer(ACode[1], Length(ACode));
  except
    on E: Exception do begin
      Msg(0, 'ExecuteCodeIndirect', E.Message);
      if F <> nil then begin
        F.Free;
        DeleteFile(FN);
      end;
      Exit;
    end;
  end;

  F.Free; // close and flush
  Result := ExecuteCode(Format('source("%s", print.eval=TRUE)', [FN]));
  DeleteFile(FN);
end;

procedure TRConsole.AssignInteger(AVariable: AnsiString; AValue: LongInt);
begin
  ExecuteCode(AVariable + ' <- ' + IntToStr(AValue));
end;

procedure TRConsole.AssignDouble(AVariable: AnsiString; AValue: Double);
begin
  ExecuteCode(AVariable + ' <- ' + FloatToStr(AValue));
end;

procedure TRConsole.AssignString(AVariable: AnsiString; AValue: AnsiString);
begin
  ExecuteCode(AVariable + ' <- "' + AValue + '"');
end;

procedure TRConsole.AssignVector(AVariable: AnsiString; AVector: PVector; ASize: LongInt);
begin
  ExecuteCodeIndirect(AVariable + ' <- ' + FormatVector(AVector, ASize));
end;

procedure TRConsole.AssignFactor(AVariable: AnsiString; AFactor: PStrVector; ASize: LongInt);
var
  i : LongInt;
  Code: AnsiString;
  dummy: TMqlStr;
begin
  Code := AVariable + ' <- factor(c(';
  ASize -= 1;
  try
    for i := 0 to ASize do begin
      Code := Code + '"' + AFactor^[i].Ptr + '"';
      if i < ASize then
        Code := Code + ',';
    end;
  except
    Msg(0, 'AssignFactor', rsErrArraySize);
    Exit;
  end;
  Code := Code + '))';
  ExecuteCodeIndirect(Code);
end;

procedure TRConsole.AssignMatrix(AVariable: AnsiString; AMatrix: PVector; ARows: LongInt; ACols: LongInt);
var
  F: TFileStream;
  FN: AnsiString;
  Dummy: Double;
begin
  try
    // try to read the last element to see whether the size is correct
    Dummy := AMatrix^[ARows * ACols - 1];
  except
    // access violation
    Msg(0, 'AssignMatrix', rsErrArraySize);
    Dummy := Dummy; // suppress the compiler hint about unused variable
    exit;
  end;

  F := nil;
  // we are usually dealing with very big matrices that contain thousands
  // of vectors of data. The fastest way to transfer them is through
  // writing the entire memory block as it is to a raw binary file and let
  // R read it with readBin() to avoid the slow formatting and parsing.
  FN := GetTempFileName(GetTempDir, 'TRConsole-matrix-' + AVariable + '-');
  FN := StringsReplace(FN, ['\'], ['/'], [rfReplaceAll]);
  try
    Msg(2, 'AssignMatrix', 'writing data');
    F := TFileStream.Create(FN, fmCreate);
    F.WriteBuffer(AMatrix^, ACols * ARows * 8); // Double is 8 byte
  except
    Msg(0, 'AssignMatrix', rsErrWriteTemp);
    if F <> nil then begin
      F.Free;
      DeleteFile(FN);
    end;
    exit;
  end;

  F.Free; // close and force flush
  ExecuteCode(Format(
    '%s <- matrix(readBin("%s", double(), %d), nrow=%d, ncol=%d, byrow=TRUE)',
    [AVariable, FN, ARows * ACols, ARows, ACols]
  ));
  DeleteFile(FN);
end;

procedure TRConsole.RowBindVector(AVariable: AnsiString; AVector: PVector; ASize: LongInt);
begin
  ExecuteCodeIndirect(AVariable + ' <- rbind(' + AVariable + ',' + FormatVector(AVector, ASize) + ')');
end;

function TRconsole.GetInteger(ACode: AnsiString): LongInt;
var
  Line : AnsiString;
begin
  if not Running then begin
    Msg(0, 'GetInteger', rsErrRNotRunning, [ACode]);
    exit(0);
  end;

  Line := ExecuteCode('as.integer(' + ACode + ')');
  if Line = 'NULL' then begin
    Msg(1, 'GetInteger', rsWarnNull);
    exit(0);
  end;

  Line := RightStr(Line, Length(Line) - Pos(']', Line) - 1);
  try
    Result := StrToInt(Line);
  except
    Msg(0, 'GetInteger', rsErrCouldNotConvertToInt, [Line]);
    Result := 0;
  end;
end;

function TRconsole.GetDouble(ACode: AnsiString): Double;
var
  Line : AnsiString;
begin
  if not Running then begin
    Msg(0, 'GetDouble', rsErrRNotRunning, [ACode]);
    exit(0);
  end;

  Line := ExecuteCode('as.double(' + ACode + ')');
  if Line = 'NULL' then begin
    Msg(1, 'GetDouble', rsWarnNull);
    exit(0);
  end;

  Line := RightStr(Line, Length(Line) - Pos(']', Line) - 1);
  Result := ConvertDouble(Line);
end;

function TRConsole.GetVector(ACode : AnsiString; AVector: PVector; ASize: LongInt): LongInt;
var
  Line : AnsiString;
  i : Integer;
  Idx, Pos1, Pos2, Len: Integer;
  ArrayLast: Integer;
begin
  if not Running then begin
    Msg(0, 'GetVector', rsErrRNotRunning, [ACode]);
    exit(0);
  end;

  ArrayLast := ASize - 1;
  Line := ExecuteCode('as.vector(' + ACode + ')');
  if Line = 'NULL' then begin
    Msg(1, 'GetVector', rsWarnNull);
    exit(0);
  end;

  Line := RightStr(Line, Length(Line) - Pos(']', Line) - 1);
  Len := Length(Line);

  // replace all line endings with space
  for i := 1 to Len do begin
    if Line[i] = #13 then Line[i] := ' '
    else if Line[i] = #10 then Line[i] := ' ';
  end;

  Idx := 0; // array index
  Pos1 := 1; // string position
  repeat
    // find next non-whitespace
    while (Pos1 < Len) and (Line[Pos1] = ' ') do Pos1 += 1;

    // new line header "[...] "
    if Line[Pos1] = '[' then begin
      while (Pos1 < Len) and (Line[Pos1] <> ' ') do Pos1 += 1;
      while (Pos1 < Len) and (Line[Pos1] = ' ') do Pos1 += 1;
    end;

    // Pos1 is now at the beginning of the number, find next withespace after it
    Pos2 := Pos1;
    while (Pos2 <= Len) and (Line[Pos2] <> ' ') do Pos2 += 1;

    if Idx <= ArrayLast then
      AVector^[Idx] := ConvertDouble(MidStr(Line, Pos1, Pos2 - Pos1));

    Pos1 := Pos2;
    Idx += 1;
  until (Pos2 > Len);

  if (Idx > ASize) then begin
    Msg(1, 'GetVector', rsWarnArrayTooSmall, [ASize, Idx]);
    Result := ASize;
  end else
  if (Idx < ASize) then begin
    Msg(1, 'GetVector', rsNoteArrayBiggerThanNeeded, [ASize, Idx]);
    Result := Idx;
  end else
    Result := Idx;
end;


end.

